import pprint

from ninja import NinjaAPI


class TestingApiEndpoints:
    tags = ["Extras"]

    def __init__(self, api: NinjaAPI):
        @api.get(
            "/extras/test/newapi1",
            tags=["Testing"]
        )
        def get_test_api_bravadosih(request):
            """
                Not for real use!
                this is kinda like bravado in use... no type hinting
            """
            if not request.user.is_superuser:
                return 403, {"message": "Hard no pall!"}

            from esi.clients import OpenAPIEsiClientProvider
            api = OpenAPIEsiClientProvider()

            return api.client.Alliance.get_alliances_alliance_id(alliance_id=1900696668)

        @api.get(
            "/extras/test/newapi2",
            tags=["Testing"]
        )
        def get_test_api_openapi(request):
            """
                Not for real use!
                this has type hinting
            """
            if not request.user.is_superuser:
                return 403, {"message": "Hard no pall!"}

            from esi.clients import OpenAPIEsiClientProvider
            api = OpenAPIEsiClientProvider()

            from esi.client.api import AllianceApi

            return AllianceApi(api.client).get_alliances_alliance_id(alliance_id=1900696668)



# Generated by Django 3.2.8 on 2022-01-09 10:19

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0063_notification_verification'),
    ]

    operations = [
        migrations.AddField(
            model_name='notification',
            name='note_text',
            field=models.ForeignKey(blank=True, default=None, null=True,
                                    on_delete=django.db.models.deletion.CASCADE, to='corptools.notificationtext'),
        ),
    ]



# Generated by Django 4.0.2 on 2022-08-06 11:44

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0074_evelocation_managed_evelocation_managed_char_and_more'),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name='notification',
            name='corptools_n_notific_77c7f2_idx',
        ),
    ]



# Generated by Django 3.1.1 on 2021-01-14 09:25

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0041_auto_20210112_1359'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='skilllist',
            name='eft',
        ),
    ]



import os

from celery import Celery

# set the default Django settings module for the 'celery' program.
os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'settings')

from django.conf import settings  # noqa

app = Celery('devauth')

# Using a string here means the worker don't have to serialize
# the configuration object to child processes.
app.config_from_object('django.conf:settings')

# Automatically try to establish the connection to the AMQP broker on
# Celery startup if it is unavailable.
app.conf.broker_connection_retry_on_startup = True

# setup priorities ( 0 Highest, 9 Lowest )
app.conf.broker_transport_options = {
    'priority_steps': list(range(10)),  # setup que to have 10 steps
    'queue_order_strategy': 'priority',  # setup que to use prio sorting
}
app.conf.task_default_priority = 5  # anything called with the task.delay() will be given normal priority (5)
app.conf.worker_prefetch_multiplier = 1  # only prefetch single tasks at a time on the workers so that prio tasks happen

app.conf.ONCE = {
    'backend': 'allianceauth.services.tasks.DjangoBackend',
    'settings': {}
}

# Load task modules from all registered Django app configs.
app.autodiscover_tasks(lambda: settings.INSTALLED_APPS)



# Generated by Django 4.2.16 on 2025-04-05 09:24

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0118_alter_skill_id'),
    ]

    operations = [
        migrations.CreateModel(
            name='CharacterAgeFilter',
            fields=[
                ('id', models.AutoField(auto_created=True,
                 primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=500)),
                ('description', models.CharField(max_length=500)),
                ('min_age', models.IntegerField(default=30)),
                ('reversed_logic', models.BooleanField(default=False,
                 help_text='If set all user with age less than the min_max_age will pass the test.')),
            ],
            options={
                'verbose_name': 'Smart Filter: Character Age Filter',
                'verbose_name_plural': 'Smart Filter: Character Age Filter',
            },
        ),
    ]



from datetime import timedelta
from typing import List, Optional

from ninja import NinjaAPI

from django.utils import timezone
from django.utils.translation import gettext as _

from corptools import models
from corptools.api.helpers import get_alts_queryset, get_main_character


class MiningApiEndpoints:

    tags = ["Account"]

    def __init__(self, api: NinjaAPI):
        @api.get(
            "account/{character_id}/mining/chars",
            response={200: dict, 403: str},
            tags=["Account"]
        )
        def get_all_characters_character_mining(request, character_id: int):
            if character_id == 0:
                character_id = request.user.profile.main_character.character_id
            response, main = get_main_character(request, character_id)

            if not response:
                return 403, _("Permission Denied")

            characters = get_alts_queryset(main)

            return characters

        @api.get(
            "account/{character_id}/mining",
            response={200: dict, 403: str},
            tags=["Account"]
        )
        def get_character_mining(request,
                                 character_id: int,
                                 filter_characters: Optional[List[int]] = [],
                                 look_back: Optional[int] = 0):
            if character_id == 0:
                character_id = request.user.profile.main_character.character_id
            response, main = get_main_character(request, character_id)

            if look_back <= 1:
                look_back = models.CorptoolsConfiguration.get_solo().aggregate_lookback

            if not response:
                return 403, _("Permission Denied")

            characters = get_alts_queryset(main)

            if len(filter_characters):
                characters = characters.exclude(
                    character_id__in=filter_characters)

            start_date = timezone.now() - timedelta(days=look_back)

            values = {}
            try:
                from moons.helpers import OreHelper
                values = OreHelper.get_ore_array_with_value()
            except Exception as e:
                pass

            mining_ledger_data = models.CharacterMiningLedger.objects.filter(
                character__character__in=characters,
                date__gte=start_date
            ).select_related(
                'character__character',
                'type_name',
                'type_name__group'
            )

            all_ores = set()
            all_groups = set()
            all_systems = set()
            t_val = 0
            t_vol = 0
            output = {}
            for t in [(timezone.now() - timedelta(days=i)).date() for i in range(look_back + 1, 0, -1)]:
                output[str(t)] = {
                    "date": str(t),
                    "ores": {},
                    "characters": {},
                    "systems": {},
                    "value": 0,
                    "volume": 0
                }

            for w in mining_ledger_data:
                _d = str(w.date)
                if _d not in output:
                    continue

                all_ores.add(w.type_name.name)
                all_groups.add(w.type_name.group.name)
                all_systems.add(w.system.name)
                vol = w.quantity * w.type_name.volume
                t_vol += vol
                value = 0

                if w.type_name.type_id in values:
                    value = w.quantity * values[w.type_name.type_id]["value"]

                t_val += value

                if w.type_name.type_id not in output[_d]["ores"]:
                    output[_d]["ores"][w.type_name.type_id] = {
                        "name": w.type_name.name,
                        "group": w.type_name.group.name,
                        "id": w.type_name.type_id,
                        "volume": 0,
                        "value": value
                    }
                output[_d]["ores"][w.type_name.type_id]["volume"] += vol
                output[_d]["ores"][w.type_name.type_id]["value"] += value
                output[_d]["volume"] += vol
                output[_d]["value"] += value

                if w.character.character.character_id not in output[_d]["characters"]:
                    output[_d]["characters"][w.character.character.character_id] = {
                        "id": w.character.character.character_id,
                        "name": w.character.character.character_name
                    }

                if w.system.system_id not in output[_d]["systems"]:
                    output[_d]["systems"][w.system.system_id] = {
                        "id": w.system.system_id,
                        "name": w.system.name
                    }

            for k, d in output.items():
                d["ores"] = list(d["ores"].values())
                d["characters"] = list(d["characters"].values())
                d["systems"] = list(d["systems"].values())

            return {
                "all_ores": list(all_ores),
                "all_groups": list(all_groups),
                "all_systems": list(all_systems),
                "total_volume": t_vol,
                "total_value": t_val,
                "data": list(output.values())
            }



# Generated by Django 2.2.12 on 2020-06-14 09:35

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0014_characteraudit_balance'),
    ]

    operations = [
        migrations.AlterField(
            model_name='skilltotals',
            name='unallocated_sp',
            field=models.IntegerField(default=None, null=True),
        ),
    ]



from typing import List

from ninja import NinjaAPI

from django.db.models import F, Q, Sum
from django.utils.translation import gettext as _

from corptools import models
from corptools.api import schema
from corptools.api.helpers import get_alts_queryset, get_main_character


class AssetsApiEndpoints:

    tags = ["Assets"]

    def __init__(self, api: NinjaAPI):

        @api.get(
            "account/{character_id}/asset/locations",
            response={200: List[schema.ValueLabel], 403: str},
            tags=self.tags
        )
        def get_character_asset_locations(request, character_id: int):
            if character_id == 0:
                character_id = request.user.profile.main_character.character_id
            response, main = get_main_character(request, character_id)

            if not response:
                return 403, _("Permission Denied")

            characters = get_alts_queryset(main)

            asset_locs = models.CharacterAsset.objects.filter(character__character__in=characters,
                                                              location_name__isnull=False).values_list('location_name').distinct()
            asset_locs = models.EveLocation.objects.filter(
                location_id__in=asset_locs).order_by('location_name')

            asset_locations = [{"label": _("Everywhere"), "value": 0},
                               {"label": _("AssetSafety"), "value": 2004}, ]
            for loc in asset_locs:
                asset_locations.append({
                    "label": loc.location_name,
                    "value": loc.location_id
                })

            return asset_locations

        @api.get(
            "account/{character_id}/asset/{location_id}/list",
            response={200: List[schema.CharacterAssetItem], 403: str},
            tags=self.tags
        )
        def get_character_asset_list(request, character_id: int, location_id: int):
            expandable_cats = [2, 6]

            if character_id == 0:
                character_id = request.user.profile.main_character.character_id
            response, main = get_main_character(request, character_id)

            if not response:
                return 403, _("Permission Denied")

            characters = get_alts_queryset(main)

            assets = models.CharacterAsset.objects\
                .filter((Q(blueprint_copy=None) | Q(blueprint_copy=False)),
                        character__character__in=characters).select_related(
                            "character", "character__character",
                            "type_name", "location_name", "type_name__group__category"
                )

            if location_id == 2004:
                asset_locations = assets.filter(
                    location_flag="AssetSafety").values_list('item_id')
                assets = assets.filter(location_id__in=asset_locations)
            elif location_id != 0:
                assets = assets.filter(Q(location_name_id=int(
                    location_id)) | Q(location_id=int(location_id)))

            output = []

            for a in assets:
                type_nm = a.type_name.name
                if a.name:
                    type_nm = f"{a.type_name.name} ({a.name})"
                loc = f"{a.location_id} ({a.location_flag})"
                if a.location_name:
                    loc = a.location_name.location_name
                output.append({
                    "character": {
                        "character_id": a.character.character.character_id,
                        "character_name": a.character.character.character_name,
                        "corporation_id": a.character.character.corporation_id,
                        "corporation_name": a.character.character.corporation_name,
                        "alliance_id": a.character.character.alliance_id,
                        "alliance_name": a.character.character.alliance_name
                    },
                    "item": {
                        "id": a.type_name.type_id,
                        "name": type_nm,
                        "cat": f"{a.type_name.group.category.name} - {a.type_name.group.name}",
                        "cat_id": a.type_name.group.category_id
                    },
                    "quantity": a.quantity,
                    "id": a.item_id,
                    "expand": True if a.type_name.group.category.category_id in expandable_cats else True if a.type_name_id == 60 else False,  # ships or asset wraps
                    "location": {
                        "id": a.location_id,
                        "name": loc
                    }
                })

            return output

        @api.get(
            "account/{character_id}/asset/{item_id}/contents",
            response={200: List[schema.CharacterAssetItem], 403: str},
            tags=self.tags
        )
        def get_character_asset_contents(request, character_id: int, item_id: int):
            if character_id == 0:
                character_id = request.user.profile.main_character.character_id
            response, main = get_main_character(request, character_id)

            if not response:
                return 403, _("Permission Denied")

            characters = get_alts_queryset(main)

            assets = models.CharacterAsset.objects\
                .filter(character__character__in=characters).select_related(
                    "character", "character__character",
                    "type_name", "location_name", "type_name__group__category"
                )
            assets = assets.filter(location_id=item_id)
            output = []

            for a in assets:
                output.append({
                    "character": {
                        "character_id": a.character.character.character_id,
                        "character_name": a.character.character.character_name,
                        "corporation_id": a.character.character.corporation_id,
                        "corporation_name": a.character.character.corporation_name,
                        "alliance_id": a.character.character.alliance_id,
                        "alliance_name": a.character.character.alliance_name
                    },
                    "item": {
                        "id": a.type_name.type_id,
                        "name": a.type_name.name,
                        "cat": f"{a.type_name.group.category.name} - {a.type_name.group.name}",
                        "cat_id": a.type_name.group.category_id

                    },
                    "quantity": a.quantity,
                    "id": a.item_id,
                    "expand": False,
                    "location": {
                        "id": item_id,
                        "name": a.location_flag
                    }
                })

            return output

        @api.get(
            "account/{character_id}/asset/{location_id}/groups",
            response={200: List[schema.CharacterAssetGroups], 403: str},
            tags=self.tags
        )
        def get_character_asset_groups(request, character_id: int, location_id: int):
            if character_id == 0:
                character_id = request.user.profile.main_character.character_id
            response, main = get_main_character(request, character_id)

            if not response:
                return 403, _("Permission Denied")

            characters = get_alts_queryset(main)

            capital_groups = [30, 547, 659, 1538, 485, 902, 513, 883, 77283]
            subcap_cat = [6]
            noteable_cats = [4, 20, 23, 25, 34, 35, 87, 91]
            structure_cats = [22, 24, 40, 41, 46, 65, 66, ]
            bpo_cats = [9]

            assets = models.CharacterAsset.objects\
                .filter((Q(blueprint_copy=None) | Q(blueprint_copy=False)),
                        character__character__in=characters)

            if location_id == 2004:
                asset_locations = assets.filter(
                    location_flag="AssetSafety").values_list('item_id')
                assets = assets.filter(location_id__in=asset_locations)
            elif location_id != 0:
                asset_locations = assets.filter(
                    location_name_id=int(location_id)).values_list('item_id')
                assets = assets.filter(Q(location_name_id=int(location_id)) | Q(
                    location_id__in=asset_locations) | Q(location_id=int(location_id)))

            assets = assets.values('type_name__group__group_id')\
                .annotate(grp_total=Sum('quantity'))\
                .annotate(grp_name=F('type_name__group__name'))\
                .annotate(grp_id=F('type_name__group_id'))\
                .annotate(cat_id=F('type_name__group__category_id'))\
                .order_by('-grp_total')

            capital_asset_groups = []
            subcap_asset_groups = []
            noteable_asset_groups = []
            structure_asset_groups = []
            bpo_asset_groups = []
            remaining_asset_groups = []

            for grp in assets:
                _grp = {
                    "label": grp['grp_name'],
                    "value": grp['grp_total'],
                }
                if grp['grp_id'] in capital_groups:
                    capital_asset_groups.append(_grp)
                elif grp['cat_id'] in subcap_cat:
                    subcap_asset_groups.append(_grp)
                elif grp['cat_id'] in noteable_cats:
                    noteable_asset_groups.append(_grp)
                elif grp['cat_id'] in structure_cats:
                    structure_asset_groups.append(_grp)
                elif grp['cat_id'] in bpo_cats:
                    bpo_asset_groups.append(_grp)
                else:
                    remaining_asset_groups.append(_grp)

            return [
                {"name": _("Capital Ships"),
                 "items": capital_asset_groups},
                {"name": _("Subcaps Ships"),
                 "items": subcap_asset_groups},
                {"name": _("Noteable Assets"),
                 "items": noteable_asset_groups},
                {"name": _("Structures"),
                 "items": structure_asset_groups},
                {"name": _("BPO"),
                 "items": bpo_asset_groups},
                {"name": _("Remaining"),
                 "items": remaining_asset_groups},
            ]



import datetime
import json
from random import random

import requests
from celery import chain as Chain, shared_task

from django.core.cache import cache
from django.core.serializers.json import DjangoJSONEncoder
from django.utils import timezone

from allianceauth.eveonline.providers import provider as eve_names
from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce

from corptools.task_helpers.housekeeping_tasks import remove_old_notifications

from .. import app_settings, providers
from ..models import (
    CharacterAsset, CharacterAudit, CharacterMarketOrder, Clone, Contract,
    CorpAsset, CorporateContract, CorporationAudit, EveItemType, EveLocation,
    EveName, JumpClone, TypePrice,
)
from ..task_helpers import corp_helpers
from ..task_helpers.update_tasks import (
    fetch_location_name, process_category_from_esi, process_map_from_esi,
    set_error_count_flag, update_ore_comp_table_from_fuzzworks,
)
from .utils import esi_error_retry, no_fail_chain

TZ_STRING = "%Y-%m-%dT%H:%M:%SZ"


logger = get_extension_logger(__name__)

# Bulk Updates


@shared_task(
    name="corptools.tasks.update_or_create_map"
)
def update_or_create_map():
    return process_map_from_esi()


@shared_task(
    name="corptools.tasks.update_ore_comp_table"
)
def update_ore_comp_table():
    return update_ore_comp_table_from_fuzzworks()


@shared_task(
    name="corptools.tasks.update_category"
)
def update_category(category_id):
    return process_category_from_esi(category_id)


@shared_task(
    name="corptools.tasks.process_ores_from_esi"
)
def process_ores_from_esi():
    return process_category_from_esi(25)


@shared_task(
    name="corptools.tasks.update_all_eve_names"
)
def update_all_eve_names(chunk=False):
    needs_update = timezone.now() - datetime.timedelta(days=30)
    en = EveName.objects.filter(last_update__lte=needs_update)
    if chunk:
        en = en[:chunk]
    for e in en:
        update_eve_name.apply_async(
            args=[e.eve_id],
            priority=7,
            countdown=random() * app_settings.CT_TASK_SPREAD_DELAY
        )


def get_error_count_flag():
    return cache.get("esi_errors_timeout", False)


@shared_task(
    bind=True,
    base=QueueOnce,
    max_retries=None,
    name="corptools.tasks.update_eve_name"
)
def update_eve_name(self, id):
    if get_error_count_flag():
        self.retry(countdown=60)

    name = EveName.objects.get(eve_id=id)
    if name.needs_update():
        try:
            if name.category == EveName.CHARACTER:
                update = eve_names.get_character(id)
                name.name = update.name
                if update.alliance:
                    alliance, _ = EveName.objects.update_or_create(
                        eve_id=update.alliance.id,
                        defaults={
                            'name': update.alliance.name,
                            'category': EveName.ALLIANCE,
                        }
                    )
                    name.alliance = alliance
                if update.corp:
                    corporation, _ = EveName.objects.update_or_create(
                        eve_id=update.corp.id,
                        defaults={
                            'name': update.corp.name,
                            'category': EveName.CORPORATION,
                        }
                    )
                    if update.alliance:
                        corporation.alliance_id = update.alliance.id
                        corporation.save()
                    name.corporation = corporation
                name.save()
            if name.category == EveName.ALLIANCE:
                update = eve_names.get_corp(id)
                name.name = update.corporation_name
                if update.alliance:
                    alliance, _ = EveName.objects.update_or_create(
                        eve_id=update.alliance.id,
                        defaults={
                            'name': update.alliance.name,
                            'category': EveName.ALLIANCE,
                        }
                    )
                    name.alliance = alliance
                name.save()
            if name.category == EveName.ALLIANCE:
                update = eve_names.get_alliance(id)
                name.name = update.name
            name.save()
        except Exception as e:  # no access
            if hasattr(e, "response"):
                if hasattr(e.response, "headers"):
                    if int(e.response.headers.get('x-esi-error-limit-remain')) < 50:
                        set_error_count_flag()
            # cooloff for a while
            name.last_updated = timezone.now()
            name.save()


@shared_task(name="corptools.tasks.process_all_categories")
def process_all_categories():
    categories = providers.esi.client.Universe.get_universe_categories().result()
    que = []

    for category in categories:
        que.append(update_category.si(category))

    Chain(que).apply_async(priority=8)

    return f"Queued {len(que)} Tasks"


@shared_task(name="corptools.tasks.run_housekeeping")
def run_housekeeping():
    notifs = remove_old_notifications()
    return notifs


@shared_task(name="corptools.tasks.update_all_raw_minerals")
def update_all_raw_minerals():
    _types = EveItemType.objects.filter(group__category_id=4)
    return update_prices_for_types(list(_types.values_list("type_id", flat=True)))


@shared_task(name="corptools.tasks.update_prices_for_types")
def update_prices_for_types(type_ids: list):
    logger.info(
        "Pulling values from Jita @`buy`-`weightedAverage`"
    )
    _str = str(type_ids.pop())
    for i in type_ids:
        _str += f",{i}"
    url = f"https://market.fuzzwork.co.uk/aggregates/?station=60003760&types={_str}"
    response = requests.get(url)
    price_data = response.json()
    price_cache = {}
    for key, item in price_data.items():
        obj, _ = EveItemType.objects.get_or_create_from_esi(key)

        ob, _ = TypePrice.objects.update_or_create(
            item=obj,
            defaults={
                "price": float(item['buy']["weightedAverage"])
            }
        )
        price_cache[obj.name] = float(item['buy']["weightedAverage"])
    return json.dumps(price_cache, indent=2)

# Bulk Updates


@shared_task(name="corptools.tasks.clear_all_etags")
def clear_all_etags():
    try:
        from django_redis import get_redis_connection
        _client = get_redis_connection("default")
    except (NotImplementedError, ModuleNotFoundError):
        from django.core.cache import caches
        default_cache = caches['default']
        _client = default_cache.get_master_client()

    keys = _client.keys(":?:etag-*")
    if keys:
        deleted = _client.delete(*keys)

    return f"Deleted {deleted} etag keys"



# Generated by Django 4.2.16 on 2025-04-08 09:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0119_characteragefilter'),
    ]

    operations = [
        migrations.AddField(
            model_name='corptoolsconfiguration',
            name='aggregate_lookback',
            field=models.IntegerField(
                default=90, help_text='Days to look back in aggregated views, like the mining ledger graphs.'),
        ),
    ]



# Generated by Django 2.2.12 on 2020-05-24 15:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0013_eveitemdogmaattribute'),
    ]

    operations = [
        migrations.AddField(
            model_name='characteraudit',
            name='balance',
            field=models.DecimalField(
                decimal_places=2, default=None, max_digits=20, null=True),
        ),
    ]



# Generated by Django 4.0.8 on 2023-02-24 09:45

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0083_skilltotalhistory'),
    ]

    operations = [
        migrations.AddField(
            model_name='characteraudit',
            name='last_update_loyaltypoints',
            field=models.DateTimeField(blank=True, default=None, null=True),
        ),
        migrations.CreateModel(
            name='LoyaltyPoint',
            fields=[
                ('id', models.AutoField(auto_created=True,
                 primary_key=True, serialize=False, verbose_name='ID')),
                ('amount', models.IntegerField(verbose_name='LP')),
                ('character', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                 to='corptools.characteraudit', verbose_name='Character')),
                ('corporation', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                 to='corptools.evename', verbose_name='NPC Corporation')),
            ],
            options={
                'verbose_name': 'Loyalty Point',
                'unique_together': {('character', 'corporation')},
            },
        ),
    ]



import bz2
from concurrent.futures import ThreadPoolExecutor, as_completed

import requests
from bravado.exception import HTTPForbidden

from django.core.cache import cache

from allianceauth.services.hooks import get_extension_logger
from esi.models import Token

from corptools import providers
from corptools.models import (
    EveItemCategory, EveItemDogmaAttribute, EveItemGroup, EveItemType,
    EveLocation, InvTypeMaterials, MapConstellation, MapRegion, MapSystem,
    MapSystemGate,
)

logger = get_extension_logger(__name__)

# from celery.utils.debug import sample_mem, memdump


def process_map_from_esi():
    # sample_mem()
    _regions = providers.esi.client.Universe.get_universe_regions().result()
    _region_models_updates = []
    _region_models_creates = []

    _constelations = []
    _constelation_model_updates = []
    _constelation_model_creates = []

    _systems = []
    _system_models_updates = []
    _system_models_creates = []

    _gate_links_array = []

    _processes = []
    _current_regions = MapRegion.objects.all().values_list('region_id', flat=True)
    with ThreadPoolExecutor(max_workers=20) as executor:
        for region in _regions:
            _processes.append(executor.submit(
                providers.esi._get_region, region, _current_regions))

    for task in as_completed(_processes):
        __region_upd, __region_new, __constelation_list = task.result()
        _constelations += __constelation_list
        if __region_upd:
            _region_models_updates.append(__region_upd)
        if __region_new:
            _region_models_creates.append(__region_new)

    MapRegion.objects.bulk_update(_region_models_updates, [
                                  'name', 'description'], batch_size=1000)  # bulk update
    MapRegion.objects.bulk_create(
        _region_models_creates, batch_size=1000, ignore_conflicts=True)  # bulk create
    # sample_mem()
    _processes = []
    _current_constellations = MapConstellation.objects.all(
    ).values_list('constellation_id', flat=True)
    with ThreadPoolExecutor(max_workers=20) as executor:
        for constellation in _constelations:
            _processes.append(executor.submit(
                providers.esi._get_constellation, constellation, _current_constellations))

    for task in as_completed(_processes):
        __constelation_upd, __constelation_new, __system_list = task.result()
        _systems += __system_list
        if __constelation_upd:
            _constelation_model_updates.append(__constelation_upd)
        if __constelation_new:
            _constelation_model_creates.append(__constelation_new)

    MapConstellation.objects.bulk_update(_constelation_model_updates, [
                                         'name', 'region_id'], batch_size=1000)  # bulk update
    MapConstellation.objects.bulk_create(
        _constelation_model_creates, batch_size=1000, ignore_conflicts=True)  # bulk create
    # sample_mem()
    _processes = []
    _current_systems = MapSystem.objects.all().values_list('system_id', flat=True)
    with ThreadPoolExecutor(max_workers=50) as executor:
        for system in _systems:
            _processes.append(executor.submit(
                providers.esi._get_system, system, _current_systems))

    for task in as_completed(_processes):
        __system_upd, __system_new, _gates = task.result()
        if __system_upd:
            _system_models_updates.append(__system_upd)
        if __system_new:
            _system_models_creates.append(__system_new)
        if _gates:
            _gate_links_array += _gates

    MapSystem.objects.bulk_update(_system_models_updates, [
                                  'name', 'constellation_id', 'star_id', 'security_class', 'x', 'y', 'z', 'security_status'], batch_size=1000)  # bulk update
    MapSystem.objects.bulk_create(
        _system_models_creates, batch_size=1000, ignore_conflicts=True)  # bulk update
    # sample_mem()
    _processes = []
    _gate_links_array = set(_gate_links_array)
    with ThreadPoolExecutor(max_workers=50) as executor:
        for gate in _gate_links_array:
            _processes.append(executor.submit(
                providers.esi._get_stargate, gate))

    _unique_gate_links = {}
    gate_models = []
    for task in as_completed(_processes):
        _to_sys, _from_sys = task.result()
        if _to_sys not in _unique_gate_links:
            _unique_gate_links[_to_sys] = []
        if _from_sys not in _unique_gate_links:
            _unique_gate_links[_to_sys].append(_from_sys)
            gate_models.append(MapSystemGate(
                from_solar_system_id=_to_sys, to_solar_system_id=_from_sys))
        elif _to_sys not in _unique_gate_links[_from_sys]:
            _unique_gate_links[_to_sys].append(_from_sys)
            gate_models.append(MapSystemGate(
                from_solar_system_id=_to_sys, to_solar_system_id=_from_sys))

    # print(_unique_gate_links)
    MapSystemGate.objects.all().delete()
    MapSystemGate.objects.bulk_create(gate_models)
    # sample_mem()
    output = "Regions: (Updated:{}, Created:{}) " \
             "Constellations: (Updated:{}, Created:{}) " \
             "Systems: (Updated:{}, Created:{}) "\
             "Gates: (Created: {}/{})".format(len(_region_models_updates),
                                              len(_region_models_creates),
                                              len(_constelation_model_updates),
                                              len(_constelation_model_creates),
                                              len(_system_models_updates),
                                              len(_system_models_creates),
                                              len(gate_models), len(_gate_links_array))
    # memdump()
    return output


def update_ore_comp_table_from_fuzzworks():
    # prime the provider or th ram will go bye bye
    providers.esi.client.Status.get_status().result()
    # Get needed SDE file
    inv_url = 'https://www.fuzzwork.co.uk/dump/latest/invTypeMaterials.csv.bz2'

    invNames_req = requests.get(inv_url)
    with open('invTypeMaterials.csv.bz2', 'wb') as iN:
        iN.write(invNames_req.content)

    # Decompress SDE files
    open('invTypeMaterials.csv', 'wb').write(
        bz2.open('invTypeMaterials.csv.bz2', 'rb').read()
    )

    # Parse file(s) and Update names object(s)
    ore_details = []  # new stuff
    mets = set()  # new stuff
    with open('invTypeMaterials.csv', encoding='UTF-8') as iN:
        csv_list = iN.read().split('\n')
        for row in csv_list[1:]:
            spl = row.split(',')
            if len(spl) > 1:
                mets.add(spl[1])
                mets.add(spl[0])
                ore_details.append(InvTypeMaterials(
                    qty=spl[2],
                    type_id=spl[0],
                    eve_type_id=spl[0],
                    material_type_id=spl[1],
                    met_type_id=spl[1]
                ))

    process_bulk_types_from_esi(mets)
    # delete it all and start again
    InvTypeMaterials.objects.all().delete()
    InvTypeMaterials.objects.bulk_create(
        ore_details, batch_size=500, ignore_conflicts=True)


def process_category_from_esi(category_id):
    _current_categories = list(
        EveItemCategory.objects.all().values_list('category_id', flat=True))
    _category, _category_creates, _groups = providers.esi._get_category(
        category_id, updates=_current_categories)

    if not _category_creates:
        EveItemCategory.objects.bulk_update([_category], ['name'])
    else:
        _category.save()

    _groups_model_updates = []
    _groups_model_creates = []

    _items = []
    _items_models_updates = []
    _items_models_creates = []
    _dogma_models_creates = []

    _processes = []
    _current_groups = list(
        EveItemGroup.objects.all().values_list('group_id', flat=True))
    with ThreadPoolExecutor(max_workers=20) as executor:
        for group in _groups:
            _processes.append(executor.submit(
                providers.esi._get_group, group, updates=_current_groups))

    for task in as_completed(_processes):
        __group, __group_new, __items_list = task.result()
        _items += __items_list
        if not __group_new:
            _groups_model_updates.append(__group)
        else:
            _groups_model_creates.append(__group)

    EveItemGroup.objects.bulk_update(
        # bulk update
        _groups_model_updates, ['name', 'category_id'], batch_size=1000)
    EveItemGroup.objects.bulk_create(
        _groups_model_creates, batch_size=1000)  # bulk create

    _processes = []
    _current_items = list(
        EveItemType.objects.all().values_list('type_id', flat=True))
    with ThreadPoolExecutor(max_workers=50) as executor:
        for item in _items:
            _processes.append(executor.submit(
                providers.esi._get_eve_type, item, updates=_current_items))

    for task in as_completed(_processes):
        __item, __item_new, __item_dogma = task.result()
        _dogma_models_creates += __item_dogma
        if not __item_new:
            _items_models_updates.append(__item)
        else:
            _items_models_creates.append(__item)

    EveItemType.objects.bulk_update(_items_models_updates,
                                    ['name', 'group_id', 'description',
                                        'mass', 'packaged_volume', 'portion_size',
                                        'volume', 'published', 'radius'], batch_size=1000)  # bulk update
    EveItemType.objects.bulk_create(
        _items_models_creates, batch_size=1000)  # bulk create

    dogma_query = EveItemDogmaAttribute.objects.filter(eve_type_id__in=_items)
    if dogma_query.exists():
        # speed and we are not caring about f-keys or signals on these models
        dogma_query._raw_delete(dogma_query.db)

    EveItemDogmaAttribute.objects.bulk_create(
        _dogma_models_creates, batch_size=1000)  # bulk create

    output = "Category ({}): " \
             "Groups: (Updated:{}, Created:{}) " \
             "Items: (Updated:{}, Created:{}, Dogma Attributes:{}".format(category_id,
                                                                          len(_groups_model_updates),
                                                                          len(_groups_model_creates),
                                                                          len(_items_models_updates),
                                                                          len(_items_models_creates),
                                                                          len(_dogma_models_creates))

    return output


def process_bulk_types_from_esi(type_ids, update_models=False):
    _items = set(type_ids)
    _items_processed = []
    _items_models_creates = []
    _items_models_updates = []
    _dogma_models_creates = []
    _groups = []
    _groups_model_creates = []
    _groups_model_updates = []
    _categories = []
    _categories_model_creates = []
    _categories_model_updates = []
    _processes = []
    _current_items = list(
        EveItemType.objects.all().values_list('type_id', flat=True))
    _current_groups = list(
        EveItemGroup.objects.all().values_list('group_id', flat=True))
    _current_categories = list(
        EveItemCategory.objects.all().values_list('category_id', flat=True))

    with ThreadPoolExecutor(max_workers=50) as executor:
        for item in _items:
            if item not in _current_items or update_models:
                _processes.append(executor.submit(
                    providers.esi._get_eve_type, item, updates=_current_items))

    for task in as_completed(_processes):
        __item, __item_new, __item_dogma = task.result()
        _items_processed.append(__item.type_id)

        if __item_new:
            _items_models_creates.append(__item)
        else:
            _items_models_updates.append(__item)
        if __item.group_id not in _current_groups or update_models:
            _groups.append(__item.group_id)

        _dogma_models_creates += __item_dogma

    _processes = []
    with ThreadPoolExecutor(max_workers=50) as executor:
        for group in set(_groups):
            if group not in _current_groups or update_models:
                _processes.append(executor.submit(
                    providers.esi._get_group, group, updates=_current_groups))
                # _current_groups.append(group)

    for task in as_completed(_processes):
        __group, __group_new, types = task.result()
        if __group_new:
            _groups_model_creates.append(__group)
        else:
            _groups_model_updates.append(__group)

        if __group.category_id not in _current_categories or update_models:
            _categories.append(__group.category_id)

    _processes = []
    with ThreadPoolExecutor(max_workers=50) as executor:
        for category in set(_categories):
            if category not in _current_categories or update_models:
                _processes.append(executor.submit(
                    providers.esi._get_category, category, updates=_current_categories))
                # _current_categories.append(category)

    for task in as_completed(_processes):
        __category, __category_new, groups = task.result()
        if __category_new:
            _categories_model_creates.append(__category)
        else:
            _categories_model_updates.append(__category)

    if len(_categories_model_creates) > 0:
        EveItemCategory.objects.bulk_create(
            _categories_model_creates, batch_size=1000, ignore_conflicts=True)  # bulk create
    if len(_categories_model_updates) > 0:
        EveItemCategory.objects.bulk_update(
            _categories_model_updates, ['name'])  # bulk update

    if len(_groups_model_creates) > 0:
        EveItemGroup.objects.bulk_create(
            _groups_model_creates, batch_size=1000, ignore_conflicts=True)  # bulk create
    if len(_groups_model_updates) > 0:
        EveItemGroup.objects.bulk_update(
            # bulk update
            _groups_model_updates, ['name', 'category_id'], batch_size=1000)

    if len(_items_models_creates) > 0:
        EveItemType.objects.bulk_create(
            _items_models_creates, batch_size=1000, ignore_conflicts=True)  # bulk create
    if len(_items_models_updates) > 0:
        EveItemType.objects.bulk_update(_items_models_updates,
                                        ['name', 'group_id', 'description',
                                         'mass', 'packaged_volume', 'portion_size',
                                         'volume', 'published', 'radius'], batch_size=1000)  # bulk update
    if len(_dogma_models_creates) > 0:
        dogma_query = EveItemDogmaAttribute.objects.filter(
            eve_type_id__in=_items_processed)
        if dogma_query.exists():
            # speed and we are not caring about f-keys or signals on these models
            dogma_query._raw_delete(dogma_query.db)

        EveItemDogmaAttribute.objects.bulk_create(
            _dogma_models_creates, batch_size=1000, ignore_conflicts=True)  # bulk create

    output = "Category: (Updated:{}, Created:{}): " \
             "Groups: (Updated:{}, Created:{}) " \
             "Items: (Updated:{}, Created:{}, Dogma Attributes:{})".format(
                 len(_categories_model_updates),
                 len(_categories_model_creates),
                 len(_groups_model_updates),
                 len(_groups_model_creates),
                 len(_items_models_updates),
                 len(_items_models_creates),
                 len(_dogma_models_creates))

    logger.debug(output)

    return True


def set_error_count_flag():
    return cache.set("esi_errors_timeout", 1, 60)


def fetch_location_name(location_id, location_flag, character_id, update=False):
    """Takes a location_id and character_id and returns a location model for items in a station/structure or in space"""

    accepted_location_flags = ['AssetSafety',
                               'Deliveries',
                               'Hangar',
                               'HangarAll',
                               'solar_system',
                               'OfficeFolder',
                               'AutoFit']

    if location_flag not in accepted_location_flags:
        if location_flag is not None:
            return None  # ship fits or in cargo holds or what ever also dont care

    existing = EveLocation.objects.filter(location_id=location_id)
    current_loc = existing.exists()

    if current_loc and location_id < 64000000:
        return existing.first()
    else:
        existing = existing.first()

    if location_id == 2004:
        # ASSET SAFETY
        return EveLocation(location_id=location_id,
                           location_name="Asset Safety")
    elif 30000000 < location_id < 33000000:  # Solar System
        system = MapSystem.objects.filter(system_id=location_id)
        if not system.exists():
            logger.error("Unknown System, Have you populated the map?")
            # TODO Do i fire the map population task here?
            return None
        else:
            system = system.first()
        return EveLocation(location_id=location_id,
                           location_name=system.name,
                           system=system)
    elif 60000000 < location_id < 64000000:  # Station ID
        station = providers.esi.client.Universe.get_universe_stations_station_id(
            station_id=location_id).result()
        system = MapSystem.objects.filter(system_id=station.get('system_id'))
        if not system.exists():
            logger.error("Unknown System, Have you populated the map?")
            # TODO Do i fire the map population task here?
            return None
        return EveLocation(location_id=location_id,
                           location_name=station.get('name'),
                           system_id=station.get('system_id'))

    req_scopes = ['esi-universe.read_structures.v1']

    token = Token.get_token(character_id, req_scopes)

    if not token:
        return None

    else:  # Structure id?
        try:
            structure = providers.esi.client.Universe.get_universe_structures_structure_id(
                structure_id=location_id, token=token.valid_access_token()).result()
        except HTTPForbidden as e:  # no access
            if int(e.response.headers.get('x-esi-error-limit-remain')) < 50:
                set_error_count_flag()
            logger.debug("Failed to get location:{}, Error:{}, Errors Remaining:{}, Time Remaining: {}".format(
                location_id, e.message, e.response.headers.get('x-esi-error-limit-remain'), e.response.headers.get('x-esi-error-limit-reset')))
            return None
        system = MapSystem.objects.filter(
            system_id=structure.get('solar_system_id'))
        if not system.exists():
            logger.error("Unknown System, Have you populated the map?")
            # TODO Do i fire the map population task here?
            return None
        if current_loc:
            existing.location_name = structure.get('name')
            return existing
        else:
            return EveLocation(location_id=location_id,
                               location_name=structure.get('name'),
                               system_id=structure.get('solar_system_id'))



from datetime import timedelta

from django.utils import timezone

from allianceauth.services.hooks import get_extension_logger

from corptools.models import Notification

logger = get_extension_logger(__name__)

NOTIFICAITON_PURGE_TYPES = [
    "AcceptedAlly",
    "AcceptedSurrender",
    "AllMaintenanceBillMsg",
    "AllWarCorpJoinedAllianceMsg",
    "AllWarDeclaredMsg",
    "AllWarInvalidatedMsg",
    "AllWarRetractedMsg",
    "AllWarSurrenderMsg",
    "AllyJoinedWarAggressorMsg",
    "AllyJoinedWarAllyMsg",
    "AllyJoinedWarDefenderMsg",
    "CorpBecameWarEligible",
    "CorpFriendlyFireDisableTimerCompleted",
    "CorpFriendlyFireDisableTimerStarted",
    "CorpFriendlyFireEnableTimerCompleted",
    "CorpFriendlyFireEnableTimerStarted",
    "CorpNoLongerWarEligible",
    "CorpWarDeclaredMsg",
    "CorpWarFightingLegalMsg",
    "CorpWarInvalidatedMsg",
    "CorpWarRetractedMsg",
    "CorpWarSurrenderMsg",
    "DeclareWar",
    "EntosisCaptureStarted",
    "InfrastructureHubBillAboutToExpire",
    "MadeWarMutual",
    "MercOfferedNegotiationMsg",
    "MercOfferRetractedMsg",
    "MutualWarInviteAccepted",
    "MutualWarInviteSent",
    "NPCStandingsGained",
    "NPCStandingsLost", "OfferedSurrender",
    "OfferedToAlly",
    "OfferToAllyRetracted",
    "OrbitalAttacked",
    "OrbitalReinforced",
    "RetractsWar",
    "SovAllClaimAquiredMsg",
    "SovAllClaimLostMsg",
    "SovCommandNodeEventStarted",
    "SovCorpClaimFailMsg",
    "SovDisruptorMsg",
    "SovereigntyIHDamageMsg",
    "SovereigntySBUDamageMsg",
    "SovereigntyTCUDamageMsg",
    "SovStationEnteredFreeport",
    "SovStructureDestroyed",
    "SovStructureReinforced",
    "StationAggressionMsg1",
    "StationAggressionMsg2",
    "StationConquerMsg",
    "StationServiceDisabled",
    "StationServiceEnabled",
    "StructureFuelAlert",
    "StructureLostArmor",
    "StructureLostShields",
    "StructureServicesOffline",
    "StructureUnderAttack",
    "StructureWentHighPower",
    "StructureWentLowPower",
    "WarAdopted",
    "WarAdopted ",
    "WarAllyInherited",
    "WarAllyOfferDeclinedMsg",
    "WarConcordInvalidates",
    "WarDeclared",
    "WarEndedHqSecurityDrop",
    "WarHQRemovedFromSpace",
    "WarInherited",
    "WarInvalid",
    "WarRetracted",
    "WarRetractedByConcord",
    "WarSurrenderDeclinedMsg",
    "WarSurrenderOfferMsg"
]


def remove_old_notifications():
    def get_qs():
        oldest = timezone.now() - timedelta(days=90)
        return Notification.objects.filter(notification_type__in=NOTIFICAITON_PURGE_TYPES,
                                           timestamp__lte=oldest)

    step = 50000

    qs_ids = list(get_qs()[:step].values_list("id", flat=True))
    Notification.objects.filter(pk__in=qs_ids).delete()
    return f"Deleted {len(qs_ids)} Notifications"



import json
import xml.etree.ElementTree as ET
from datetime import timedelta

from celery.schedules import crontab
from django_celery_beat.models import CrontabSchedule, PeriodicTask

from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.core.exceptions import ObjectDoesNotExist, PermissionDenied
from django.shortcuts import redirect, render
from django.template import TemplateDoesNotExist
from django.urls import reverse
from django.utils import timezone

from allianceauth.crontab.utils import offset_cron
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger
from esi.decorators import _check_callback, token_required
from esi.views import sso_redirect

from . import __version__, app_settings
from .api.corporation import dashboards
from .forms import UploadForm
from .models import (
    CharacterAudit, CorpAsset, CorporationAudit, CorptoolsConfiguration,
    EveItemCategory, EveItemDogmaAttribute, EveItemGroup, EveItemType,
    EveLocation, EveName, InvTypeMaterials, MapConstellation, MapJumpBridge,
    MapRegion, MapSystem, MapSystemGate, SkillList, Structure,
)
from .tasks import (
    check_account, clear_all_etags, update_all_characters, update_all_corps,
    update_all_eve_names, update_all_locations, update_character, update_corp,
    update_or_create_map,
)

logger = get_extension_logger(__name__)

CORP_REQUIRED_SCOPES = [

    # Tracking
    'esi-corporations.track_members.v1',
    'esi-corporations.read_titles.v1',
    'esi-corporations.read_corporation_membership.v1',
    'esi-killmails.read_corporation_killmails.v1',

    # Moons
    'esi-industry.read_corporation_mining.v1',

    # Structures
    'esi-planets.read_customs_offices.v1',
    'esi-corporations.read_starbases.v1',
    'esi-corporations.read_structures.v1',

    # Wallets
    'esi-wallet.read_corporation_wallets.v1',
    'esi-markets.read_corporation_orders.v1',
    'esi-industry.read_corporation_jobs.v1',
    'esi-corporations.read_divisions.v1',

    # Assets
    'esi-assets.read_corporation_assets.v1',

    # All...
    'esi-search.search_structures.v1',
    'esi-universe.read_structures.v1',
    'esi-characters.read_corporation_roles.v1',

]


@login_required
@token_required(scopes=app_settings.get_character_scopes())
def add_char(request, token):
    CharacterAudit.objects.update_or_create(
        character=EveCharacter.objects.get_character_by_id(token.character_id))
    update_character.apply_async(args=[token.character_id], kwargs={
                                 "force_refresh": True}, priority=6)
    return redirect('corptools:reactmain', character_id=token.character_id)


@login_required
@token_required(scopes=CORP_REQUIRED_SCOPES)
def add_corp(request, token):
    char = EveCharacter.objects.get_character_by_id(token.character_id)
    corp, created = EveCorporationInfo.objects.get_or_create(corporation_id=char.corporation_id,
                                                             defaults={'member_count': 0,
                                                                       'corporation_ticker': char.corporation_ticker,
                                                                       'corporation_name': char.corporation_name
                                                                       })
    CorporationAudit.objects.update_or_create(corporation=corp)
    update_all_corps.apply_async(priority=6)
    return redirect('corptools:corp_react')


@login_required
def add_corp_section(request, *args, **kwargs):

    tracking = request.GET.get('t', False)
    assets = request.GET.get('a', False)
    structures = request.GET.get('s', False)
    starbases = request.GET.get('sb', False)
    wallets = request.GET.get('w', False)
    moons = request.GET.get('m', False)
    pocos = request.GET.get('p', False)
    contracts = request.GET.get('c', False)

    # if we're coming back from SSO with a new token, return it
    token = _check_callback(request)
    if token:
        logger.debug(
            "Got new token from %s session %s. Returning to view.",
            request.user,
            request.session.session_key[:5]
        )
        char = EveCharacter.objects.get_character_by_id(token.character_id)
        corp, created = EveCorporationInfo.objects.get_or_create(corporation_id=char.corporation_id,
                                                                 defaults={'member_count': 0,
                                                                           'corporation_ticker': char.corporation_ticker,
                                                                           'corporation_name': char.corporation_name
                                                                           })
        CorporationAudit.objects.update_or_create(corporation=corp)
        update_corp.apply_async(priority=6, args=[char.corporation_id])

        return redirect("{}#status".format(reverse('corptools:corp_react')))

    scopes = app_settings._corp_scopes_base

    if tracking:
        scopes += app_settings._corp_scopes_tracking

    if structures:
        scopes += app_settings._corp_scopes_structures

    if starbases:
        scopes += app_settings._corp_scopes_starbases

    if moons:
        scopes += app_settings._corp_scopes_moons

    if wallets:
        scopes += app_settings._corp_scopes_wallets

    if assets and not pocos:
        scopes += app_settings._corp_scopes_assets

    if pocos:
        scopes += app_settings._corp_scopes_pocos
        scopes += app_settings._corp_scopes_assets

    if contracts:
        scopes += app_settings._corp_scopes_contracts
    # user has selected to add a new token
    return sso_redirect(request, scopes=scopes)


@login_required
@permission_required('corptools.view_characteraudit')
def corptools_menu(request):
    # get available models
    cas = CharacterAudit.objects.visible_to(request.user)\
        .select_related('character__character_ownership__user__profile__main_character',
                        'character__characteraudit')\
        .prefetch_related('character__character_ownership__user__character_ownerships')\
        .prefetch_related('character__character_ownership__user__character_ownerships__character')\
        .prefetch_related('character__character_ownership__user__character_ownerships__character__characteraudit')\


    chars = {}
    orphans = []
    for char in cas:
        try:
            main = char.character.character_ownership.user.profile.main_character
            if main:
                if main.character_name not in chars:
                    chars[str(main.character_id)] = {'main': main,
                                                     'audit': char}
            else:
                orphans.append(char)
        except ObjectDoesNotExist:
            orphans.append(char)

    if len(chars) == 1:
        return redirect('corptools:overview', chars[list(chars.keys())[0]]['main'].character_id)

    return render(request, 'corptools/menu.html', context={'characters': chars, 'orphans': orphans})


@login_required
@permission_required('corptools.admin')
def admin(request):
    # get available models
    names = EveName.objects.all().count()
    types = EveItemType.objects.all().count()
    dogma = EveItemDogmaAttribute.objects.all().count()
    groups = EveItemGroup.objects.all().count()
    categorys = EveItemCategory.objects.all().count()
    type_mets = InvTypeMaterials.objects.count()
    regions = MapRegion.objects.all().count()
    constellations = MapConstellation.objects.all().count()
    systems = MapSystem.objects.all().count()
    gates = MapSystemGate.objects.all().count()
    location = EveLocation.objects.all().count()
    bridges = MapJumpBridge.objects.all().count()

    characters = CharacterAudit.objects.all().count()
    actives = CharacterAudit.get_oldest_qs().count()
    skilllists = SkillList.objects.all().count()
    corpations = CorporationAudit.objects.all().count()

    char_tasks = PeriodicTask.objects.filter(
        task='corptools.tasks.update_subset_of_characters', enabled=True).count()
    corp_tasks = PeriodicTask.objects.filter(
        task='corptools.tasks.update_all_corps', enabled=True).count()

    context = {
        "names": names,
        "types": types,
        "dogma": dogma,
        "groups": groups,
        "categorys": categorys,
        "characters": characters,
        "active_chars": actives,
        "skilllists": skilllists,
        "corpations": corpations,
        "type_mets": type_mets,
        "regions": regions,
        "constellations": constellations,
        "systems": systems,
        "location": location,
        "bridges": bridges,
        "gates": gates,
        "char_tasks": char_tasks,
        "corp_tasks": corp_tasks,
        "form": UploadForm(),
        "ct_config": CorptoolsConfiguration.get_solo(),
        "app_settings": app_settings
    }

    return render(request, 'corptools/admin.html', context=context)


@login_required
@permission_required('corptools.admin')
def admin_run_tasks(request):
    if request.method == 'POST':
        if request.POST.get('run_universe'):
            messages.info(request, "Queued update_or_create_map")
            update_or_create_map.apply_async(priority=6)
        if request.POST.get('run_update_all'):
            messages.info(request, "Queued update_all_characters")
            update_all_characters.apply_async(priority=6)
        if request.POST.get('run_update_eve_models'):
            messages.info(request, "Queued update_all_eve_names")
            update_all_eve_names.apply_async(priority=6)
        if request.POST.get('run_corp_updates'):
            messages.info(request, "Queued update_all_corps")
            update_all_corps.apply_async(
                priority=6, kwargs={"force_refresh": True})
        if request.POST.get('run_locations'):
            messages.info(request, "Queued update_all_locations")
            update_all_locations.apply_async(priority=6)
        if request.POST.get('run_clear_etag'):
            messages.info(request, "Queued clear_all_etags")
            clear_all_etags.apply_async(priority=1)

    return redirect('corptools:admin')


@login_required
@permission_required('corptools.admin')
def admin_create_tasks(request):

    schedule_a = CrontabSchedule.from_schedule(
        offset_cron(crontab(minute='15,45')))
    schedule_char, created = CrontabSchedule.objects.get_or_create(
        minute=schedule_a.minute,
        hour=schedule_a.hour,
        day_of_month=schedule_a.day_of_month,
        month_of_year=schedule_a.month_of_year,
        day_of_week=schedule_a.day_of_week,
        timezone=schedule_a.timezone,
    )

    schedule_b = CrontabSchedule.from_schedule(
        offset_cron(crontab(minute='30')))
    schedule_corp, schedule_corp_created = CrontabSchedule.objects.get_or_create(
        minute=schedule_b.minute,
        hour=schedule_b.hour,
        day_of_month=schedule_b.day_of_month,
        month_of_year=schedule_b.month_of_year,
        day_of_week=schedule_b.day_of_week,
        timezone=schedule_b.timezone,
    )

    PeriodicTask.objects.update_or_create(
        task='corptools.tasks.update_subset_of_characters',
        defaults={
            'crontab': schedule_char,
            'name': 'Character Audit Rolling Update',
            'enabled': True
        }
    )

    PeriodicTask.objects.update_or_create(
        task='corptools.tasks.update_all_corps',
        defaults={
            'crontab': schedule_corp,
            'name': 'Corporation Audit Update',
            'enabled': True
        }
    )

    # https://github.com/celery/django-celery-beat/issues/106
    messages.info(
        request, "Created/Reset Character and Corporation Task to defaults")

    return redirect('corptools:admin')


@login_required
@permission_required('corptools.admin')
def admin_add_pyfa_xml(request):
    if request.method == 'POST':
        form = UploadForm(request.POST, request.FILES)
        if form.is_valid():
            xml_file = form.files.get('file')
            if xml_file.content_type == 'text/xml':
                skills = {}
                tree = ET.parse(xml_file)
                root = tree.getroot()
                if root.tag == "plan":
                    for child in root:
                        if child.tag == "entry":
                            skill = child.attrib.get('skill')
                            level = child.attrib.get('level')
                            if skill in skills:
                                if level > skills[skill]:
                                    skills[skill] = level
                            else:
                                skills[skill] = level
                sl, created = SkillList.objects.update_or_create(name=request.POST["name"],
                                                                 defaults={
                    "skill_list": json.dumps(skills)
                })

                messages.success(request, "File Uploaded and Processed! {}: {}".format(
                    ("Created" if created else "Updated"),
                    request.POST["name"]
                ))
            else:
                messages.error(
                    request, "File Upload Failed! Must be XML Exported from PYFA!")
        else:
            messages.error(request, "File Upload Failed! Invalid Form")
    else:
        messages.error(
            request, "File Upload Failed! What are you doing your not meant to be here?")

    return redirect('corptools:admin')


@login_required
@permission_required('corptools.admin')
def skill_list_editor(request):
    pass


@login_required
def update_account(request, character_id):
    check_account.apply_async(args=[character_id], priority=6)
    messages.success(request, "Requested an update task.")
    return redirect('corptools:view')


@login_required
def react_menu(request):
    return redirect('corptools:reactlegacymain', request.user.profile.main_character.character_id)


@login_required
def react_menu_v4(request):
    if request.path.startswith('/audit/r_beta/corp/glance'):
        return render(request, 'corptools/base_bs5.html', context={"version": __version__, "app_name": "corptools/bs5", "page_title": "Audit"})
    return redirect('corptools:reactmain', request.user.profile.main_character.character_id)


@login_required
def react_corp_beta(request):
    return render(request, 'corptools/base_bs5.html', context={"version": __version__, "app_name": "corptools/bs5", "page_title": "Audit"})


@login_required
def react_main(request, character_id):
    try:
        return render(request, 'corptools/base_bs5.html', context={"version": __version__, "app_name": "corptools/bs5", "page_title": "Audit"})
    except TemplateDoesNotExist:
        pass

    return render(request, 'corptools/character/react_base.html', context={"version": __version__, "app_name": "corptools/char", "page_title": "Character Audit"})


@login_required
def v3_ui_render(request, character_id):
    return render(request, 'corptools/character/react_base.html', context={"version": __version__, "app_name": "corptools/char", "page_title": "Character Audit"})


@login_required
def react_corp(request):
    return render(request, 'corptools/corporation/react_base.html', context={"version": __version__, "app_name": "corptools/corp", "page_title": "Corporation Audit"})


@login_required
def fuel_levels(request):
    if not (request.user.has_perm('corptools.own_corp_manager')
            or request.user.has_perm('corptools.alliance_corp_manager')
            or request.user.has_perm('corptools.state_corp_manager')
            or request.user.has_perm('corptools.global_corp_manager')
            or request.user.has_perm('corptools.holding_corp_structures')):
        raise PermissionDenied("No perms to view")

    # hourly fuel reqs [ cit, eng, ref, flex ]
    citadel_service_mods = {
        'Clone Bay': [7.5, 10, 10, 10],
        'Market': [30, 40, 40, 40],
        'Manufacturing (Capitals)': [24, 18, 24, 24],
        # how to detect this
        'Standup Hyasyoda Research Lab': [10, 7.5, 10, 10],
        'Invention': [12, 9, 12, 12],
        'Manufacturing (Standard)': [12, 9, 12, 12],
        'Blueprint Copying': [12, 9, 12, 12],
        'Material Efficiency Research': [0, 0, 0, 0],  # part of above
        'Time Efficiency Research': [0, 0, 0, 0],  # part of above
        'Manufacturing (Super Capitals)': [36, 27, 36, 36],
        'Composite Reactions': [15, 15, 12, 15],
        'Hybrid Reactions': [15, 15, 12, 15],
        'Moon Drilling': [5, 5, 5, 5],
        'Biochemical Reactions': [15, 15, 12, 15],
        'Reprocessing': [10, 10, 7.5, 10],
        'Jump Access': [0, 0, 0, 15],  # large to show errors
        'Cynosural Jammer': [0, 0, 0, 40],
        'Jump Gate Access': [0, 0, 0, 30],
        'Jump Bridge Access': [0, 0, 0, 30],
        'Automatic Moon Drilling': [0, 0, 0, 5]
    }

    cit = [35833, 47516, 47512, 47513, 47514, 47515, 35832, 35834]
    eng = [35827, 35826, 35825]
    ref = [35835, 35836]
    fle = [37534, 35841, 35840, 81826]

    flex_fuel_types = [81143]

    all_structures = Structure.get_visible(request.user).select_related(
        'corporation', 'corporation__corporation', 'system_name', 'type_name',
        'system_name__constellation', 'system_name__constellation__region'
    ).prefetch_related('structureservice_set')

    structure_tree = []
    total_hourly_fuel = 0
    for s in all_structures:
        structure_hourly_fuel = 0
        structure_type = 99

        if s.type_id in cit:
            structure_type = 0
        elif s.type_id in eng:
            structure_type = 1
        elif s.type_id in ref:
            structure_type = 2
        elif s.type_id in fle:
            structure_type = 3

        for service in s.structureservice_set.all():
            if service.state == 'online':
                fuel_use = citadel_service_mods[service.name][structure_type]
                total_hourly_fuel += fuel_use
                structure_hourly_fuel += fuel_use

        hours = 0

        if s.fuel_expires is not None:
            hours = (s.fuel_expires - timezone.now()).total_seconds() // 3600

        extras = None

        if s.type_id == 81826:
            fuels = CorpAsset.objects.filter(
                type_id__in=flex_fuel_types,
                location_id=s.structure_id
            ).select_related(
                'type_name'
            )
            out = {
                "qty": 0,
            }
            for itm in fuels:
                out["name"] = itm.type_name.name
                out["qty"] += itm.quantity

            out["expires"] = timezone.now() + timedelta(hours=out["qty"] / 150)
            if out['qty'] > 0:
                extras = out

        structure_tree.append(
            {
                'structure': s,
                'fuel_req': structure_hourly_fuel,
                "current_blocks": int(hours * structure_hourly_fuel),
                "extra_fuel_info": extras,
                "90day": max(
                    int(
                        (
                            structure_hourly_fuel * 90 * 24
                        ) - (
                            hours * structure_hourly_fuel
                        )
                    ),
                    0
                )
            }
        )

    context = {
        'structures': structure_tree,
        'total_hourly_fuel': total_hourly_fuel,
    }
    return render(request, 'corptools/dashboards/fuel_dash.html', context=context)


@login_required
def metenox_levels(request):
    if not (request.user.has_perm('corptools.own_corp_manager')
            or request.user.has_perm('corptools.alliance_corp_manager')
            or request.user.has_perm('corptools.state_corp_manager')
            or request.user.has_perm('corptools.global_corp_manager')
            or request.user.has_perm('corptools.holding_corp_structures')):
        raise PermissionDenied("No perms to view")

    context = {
        'structures': dashboards.DashboardApiEndpoints.get_dashboard_drills_levels(request),
    }
    return render(request, 'corptools/dashboards/metenox_dash.html', context=context)



# Generated by Django 4.2.16 on 2025-03-07 09:24

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0116_starbase'),
    ]

    operations = [
        migrations.AddField(
            model_name='starbase',
            name='name',
            field=models.CharField(
                blank=True, default='', max_length=255, null=True),
        ),
    ]



# Generated by Django 2.2.12 on 2020-08-28 08:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0030_charactermarketorder_corporationmarketorder'),
    ]

    operations = [
        migrations.AddField(
            model_name='characteraudit',
            name='active',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='characteraudit',
            name='cache_expire_skill_que',
            field=models.DateTimeField(blank=True, default=None, null=True),
        ),
        migrations.AddField(
            model_name='characteraudit',
            name='cache_expire_wallet',
            field=models.DateTimeField(blank=True, default=None, null=True),
        ),
        migrations.AddField(
            model_name='characteraudit',
            name='last_update_skill_que',
            field=models.DateTimeField(blank=True, default=None, null=True),
        ),
        migrations.AddField(
            model_name='characteraudit',
            name='last_update_wallet',
            field=models.DateTimeField(blank=True, default=None, null=True),
        ),
    ]



# Generated by Django 4.0.8 on 2023-02-01 09:18

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0082_characteraudit_last_update_contracts_contract_and_more'),
    ]

    operations = [
        migrations.CreateModel(
            name='SkillTotalHistory',
            fields=[
                ('id', models.AutoField(auto_created=True,
                 primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateTimeField(auto_now=True)),
                ('total_sp', models.BigIntegerField()),
                ('unallocated_sp', models.IntegerField(default=0)),
                ('character', models.ForeignKey(
                    on_delete=django.db.models.deletion.CASCADE, to='corptools.characteraudit')),
            ],
        ),
    ]



from typing import List

from ninja import NinjaAPI

from django.utils.translation import gettext as _

from corptools import models
from corptools.api import schema
from corptools.api.helpers import get_alts_queryset, get_main_character


class RolesApiEndpoints:

    tags = ["Account"]

    def __init__(self, api: NinjaAPI):
        @api.get(
            "account/{character_id}/roles",
            response={200: List[schema.CharacterRoles], 403: str},
            tags=self.tags
        )
        def get_character_roles(request, character_id: int):
            if character_id == 0:
                character_id = request.user.profile.main_character.character_id
            response, main = get_main_character(request, character_id)

            if not response:
                return 403, _("Permission Denied")

            characters = get_alts_queryset(main)

            roles_data = models.CharacterRoles.objects\
                .filter(character__character__in=characters)\
                .select_related('character__character').prefetch_related('titles')

            output = []
            for r in roles_data:
                titles = []
                for t in r.titles.all():
                    titles.append({
                        "id": t.title_id,
                        "name": t.title
                    })
                output.append({
                    "character": r.character.character,
                    "director": r.director,
                    "station_manager": r.station_manager,
                    "personnel_manager": r.personnel_manager,
                    "accountant": r.accountant,
                    "titles": titles
                })
            return output



from datetime import timedelta
from typing import List, Optional

from ninja import NinjaAPI

from django.utils import timezone
from django.utils.translation import gettext as _

from allianceauth.services.hooks import get_extension_logger

from corptools import models
from corptools.api.helpers import (
    get_alts_queryset, get_corporation_characters, get_main_character,
)

from ..schema import Character

logger = get_extension_logger(__name__)


class MiningApiEndpoints:

    tags = ["Corporation"]

    def __init__(self, api: NinjaAPI):
        @api.get(
            "corporation/{corporation_id}/mining",
            response={200: dict, 403: str},
            tags=self.tags
        )
        def get_corporation_agregated_mining(
            request,
            corporation_id: int,
            look_back: Optional[int] = 0
        ):
            perms = (
                request.user.has_perm('corptools.own_corp_manager')
                | request.user.has_perm('corptools.alliance_corp_manager')
                | request.user.has_perm('corptools.state_corp_manager')
                | request.user.has_perm('corptools.global_corp_manager')
            )

            if not perms:
                logger.error(
                    f"Permission Denied for {request.user} to view mining!")
                return 403, "Permission Denied!"

            characters = get_corporation_characters(request, corporation_id)

            if look_back <= 1:
                look_back = models.CorptoolsConfiguration.get_solo().aggregate_lookback

            start_date = timezone.now() - timedelta(days=look_back)

            values = {}
            try:
                from moons.helpers import OreHelper
                values = OreHelper.get_ore_array_with_value()
            except Exception as e:
                pass

            mining_ledger_data = models.CharacterMiningLedger.objects.filter(
                character__character__in=characters,
                date__gte=start_date
            ).select_related(
                'character__character',
                'character__character__character_ownership',
                'character__character__character_ownership__user',
                'character__character__character_ownership__user__profile',
                'character__character__character_ownership__user__profile__main_character',
                'type_name',
                'type_name__group'
            )

            all_ores = set()
            all_groups = set()
            all_systems = set()
            t_val = 0
            t_vol = 0
            output = {}
            for t in [(timezone.now() - timedelta(days=i)).date() for i in range(look_back + 1, 0, -1)]:
                output[str(t)] = {
                    "date": str(t),
                    "ores": {},
                    "characters": {},
                    "systems": {},
                    "value": 0,
                    "volume": 0
                }

            for w in mining_ledger_data:
                _d = str(w.date)
                if _d not in output:
                    continue
                all_ores.add(w.type_name.name)
                all_groups.add(w.type_name.group.name)
                all_systems.add(w.system.name)
                vol = w.quantity * w.type_name.volume
                t_vol += vol
                value = 0

                if w.type_name.type_id in values:
                    value = w.quantity * values[w.type_name.type_id]["value"]

                t_val += value

                if w.type_name.type_id not in output[_d]["ores"]:
                    output[_d]["ores"][w.type_name.type_id] = {
                        "name": w.type_name.name,
                        "group": w.type_name.group.name,
                        "id": w.type_name.type_id,
                        "volume": 0,
                        "value": value
                    }
                output[_d]["ores"][w.type_name.type_id]["volume"] += vol
                output[_d]["ores"][w.type_name.type_id]["value"] += value
                output[_d]["volume"] += vol
                output[_d]["value"] += value

                try:
                    mc = w.character.character.character_ownership.user.profile.main_character
                except Exception as e:
                    mc = w.character.character

                if mc.character_id not in output[_d]["characters"]:
                    output[_d]["characters"][mc.character_id] = {
                        "id": mc.character_id,
                        "name": mc.character_name
                    }

                if w.system.system_id not in output[_d]["systems"]:
                    output[_d]["systems"][w.system.system_id] = {
                        "id": w.system.system_id,
                        "name": w.system.name
                    }

            for k, d in output.items():
                d["ores"] = list(d["ores"].values())
                d["characters"] = list(d["characters"].values())
                d["systems"] = list(d["systems"].values())

            return {
                "all_ores": list(all_ores),
                "all_groups": list(all_groups),
                "all_systems": list(all_systems),
                "total_volume": t_vol,
                "total_value": t_val,
                "data": list(output.values())
            }



import datetime
from functools import wraps

from bravado.exception import HTTPError
from celery import signature
from celery.exceptions import Retry
from celery_once import AlreadyQueued

from django.core.cache import cache
from django.utils import timezone

from allianceauth.services.hooks import get_extension_logger

logger = get_extension_logger(__name__)


def enqueue_next_task(chain, delay=1):
    """
        Queue next task, and attach the rest of the chain to it.
    """
    while (len(chain)):
        _t = chain.pop(0)
        _t = signature(_t)
        _t.kwargs.update({"chain": chain})
        try:
            _t.apply_async(priority=9, countdown=delay)
        except AlreadyQueued:
            # skip this task as it is already in the queue
            logger.warning(f"Skipping task as its already queued {_t}")
            continue
        break


def set_error_flag(timeout):
    tout = timezone.now() + datetime.timedelta(seconds=timeout)
    cache.set("esi_error_timeout", tout, timeout=timeout + 1)


def get_error_flag():
    return cache.get("esi_error_timeout", default=timezone.now())


def clear_error_flag():
    cache.delete("esi_error_timeout")


def esi_error_retry(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        _ret = None

        if get_error_flag() >= timezone.now():
            logger.warning("Hit ESI error limit! will retry tasks!")
            args[0].retry(countdown=61)
        else:
            clear_error_flag()
        try:
            _ret = func(*args, **kwargs)
        except Exception as e:
            if isinstance(e, (HTTPError)):
                code = e.status_code
                if code == 420:
                    logger.warning(f"Hit ESI error limit! Pausing Tasks! {e}")
                    set_error_flag(60)
                    args[0].retry(countdown=61)
            elif isinstance(e, (OSError)):
                logger.warning(f"Hit ESI error limit! Pausing Tasks! {e}")
            raise e
        return _ret
    return wrapper


def no_fail_chain(func):
    """
        Decorator to chain tasks provided in the chain kwargs regardless of task failures.
        Be sure to add chain=[] to your kwargs. TODO make this not needed.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        excp = None
        _ret = None
        try:
            _ret = func(*args, **kwargs)
        except Exception as e:
            excp = e
        finally:
            _chn = kwargs.get("chain", [])
            if not isinstance(excp, Retry):
                enqueue_next_task(_chn)
            if excp:
                raise excp
        return _ret
    return wrapper



from django.core.management.base import BaseCommand

from corptools.models import Notification
from corptools.task_helpers import sanitize_notification_type


class Command(BaseCommand):
    help = 'Update any known notifications that have "unknown type (####)"'

    def handle(self, *args, **options):
        knowns = Notification.objects.filter(
            notification_type__icontains="unknown notification type"
        )
        self.stdout.write(f"Found {knowns.count()} notifications that need updating")

        bad_types = []
        fixed_types = {}

        for note in knowns:
            new_type = sanitize_notification_type(note.notification_type)
            if new_type != note.notification_type:
                if new_type not in fixed_types:
                    fixed_types[new_type] = 0
                note.notification_type = new_type
                note.save()
                fixed_types[new_type] += 1
            else:
                if new_type not in bad_types:
                    self.stdout.write(
                        f"New Unknown Type '{new_type}'. Unable to update"
                    )
                    bad_types.append(new_type)

        self.stdout.write(
            f"Complete: {fixed_types}"
        )



from django.core.management.base import BaseCommand

from corptools.models import EveLocation, Structure


class Command(BaseCommand):
    help = 'Convert All Strucures to EveLocations'

    def handle(self, *args, **options):
        for s in Structure.objects.all():
            EveLocation.objects.update_or_create(
                location_id=s.structure_id,
                defaults={
                    "location_name": s.name,
                    "system_id": s.system_name.system_id
                }
            )



# Generated by Django 4.2.4 on 2023-08-02 22:59

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0089_highestspfilter'),
    ]

    operations = [
        migrations.RenameField(
            model_name='corporationaudit',
            old_name='cache_expire_assets',
            new_name='last_change_assets',
        ),
        migrations.RenameField(
            model_name='corporationaudit',
            old_name='cache_expire_moons',
            new_name='last_change_moons',
        ),
        migrations.RenameField(
            model_name='corporationaudit',
            old_name='cache_expire_observers',
            new_name='last_change_observers',
        ),
        migrations.RenameField(
            model_name='corporationaudit',
            old_name='cache_expire_pub_data',
            new_name='last_change_pub_data',
        ),
        migrations.RenameField(
            model_name='corporationaudit',
            old_name='cache_expire_structures',
            new_name='last_change_structures',
        ),
        migrations.RenameField(
            model_name='corporationaudit',
            old_name='cache_expire_wallet',
            new_name='last_change_wallet',
        ),
    ]



import datetime
import json

from celery import shared_task

from django.core.cache import cache
from django.core.serializers.json import DjangoJSONEncoder
from django.utils import timezone

from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce

from ..models import (
    CharacterAsset, CharacterAudit, CharacterMarketOrder, Clone, Contract,
    CorpAsset, CorporateContract, EveLocation, JumpClone,
)
from ..task_helpers.update_tasks import fetch_location_name
from .utils import esi_error_retry, no_fail_chain

TZ_STRING = "%Y-%m-%dT%H:%M:%SZ"


logger = get_extension_logger(__name__)

# Bulk Updates


def build_location_cache_tag(location_id):
    return f"loc_id_{location_id}"


def build_location_cooloff_cache_tag(location_id):
    return f"cooldown_loc_id_{location_id}"


def get_location_cooloff(location_id):
    return cache.get(build_location_cooloff_cache_tag(location_id), False)


def set_location_cooloff(location_id):
    # timeout for 7 days
    return cache.set(build_location_cooloff_cache_tag(location_id), True, (60 * 60 * 24 * 7))


def get_error_count_flag():
    return cache.get("esi_errors_timeout", False)


def location_get(location_id):
    cache_tag = build_location_cache_tag(location_id)
    data = json.loads(cache.get(cache_tag, '{"date":false, "characters":[]}'))
    if data.get('date') is not False:
        try:
            data['date'] = datetime.datetime.strptime(
                data.get('date'), TZ_STRING).replace(tzinfo=timezone.utc)
        except Exception:
            data['date'] = datetime.datetime.min.replace(tzinfo=timezone.utc)
    return data


CACHE_TIMEOUT = 60*60*24*30


def location_set(location_id, character_id):
    cache_tag = build_location_cache_tag(location_id)
    date = timezone.now() - datetime.timedelta(days=7)
    data = location_get(location_id)
    if data.get('date') is not False:
        if data.get('date') > date:
            data.get('characters').append(character_id)
            cache.set(cache_tag, json.dumps(
                data, cls=DjangoJSONEncoder), CACHE_TIMEOUT)
            return True
        else:
            data['date'] = timezone.now().strftime(TZ_STRING)
            data['characters'] = [character_id]
            cache.set(cache_tag, json.dumps(
                data, cls=DjangoJSONEncoder), CACHE_TIMEOUT)

    if character_id not in data.get('characters'):
        data.get('characters').append(character_id)
        data['date'] = timezone.now().strftime(TZ_STRING)
        cache.set(cache_tag, json.dumps(
            data, cls=DjangoJSONEncoder), CACHE_TIMEOUT)
        return True

    return False


@shared_task(
    bind=True,
    base=QueueOnce,
    max_retries=None,
    name="corptools.tasks.update_citadel_names"
)
@esi_error_retry
def update_citadel_names(self):
    citadels = EveLocation.objects.filter(location_id__gte=64000000)
    for c in citadels:
        update_location.apply_async(
            args=[c.location_id],
            kwargs={"force_citadel": True},
            priority=7
        )


def get_character_lists(location_id):
    cached_data = location_get(location_id)

    date = timezone.now() - datetime.timedelta(days=7)

    asset = CharacterAsset.objects.filter(
        location_id=location_id).select_related('character__character')
    clone = Clone.objects.filter(
        location_id=location_id).select_related('character__character')
    jumpclone = JumpClone.objects.filter(
        location_id=location_id).select_related('character__character')
    marketorder = CharacterMarketOrder.objects.filter(
        location_id=location_id).select_related('character__character')

    if cached_data.get('date') is not False:
        if cached_data.get('date') > date:
            asset = asset.exclude(
                character__character__character_id__in=cached_data.get('characters'))
            clone = clone.exclude(
                character__character__character_id__in=cached_data.get('characters'))
            jumpclone = jumpclone.exclude(
                character__character__character_id__in=cached_data.get('characters'))
            marketorder = marketorder.exclude(
                character__character__character_id__in=cached_data.get('characters'))

    # location_flag = None
    char_ids = []

    if asset.exists():
        char_ids += list(asset.values_list('character__character__character_id', flat=True))
    if clone.exists():
        char_ids += list(clone.values_list('character__character__character_id', flat=True))
    if jumpclone.exists():
        char_ids += list(jumpclone.values_list(
            'character__character__character_id', flat=True))
    if marketorder.exists():
        char_ids += list(marketorder.values_list(
            'character__character__character_id', flat=True))

    return set(char_ids)


def update_missing_locations(location_id):
    count = CharacterAsset.objects.filter(
        location_id=location_id,
        location_name__isnull=True
    ).update(location_name_id=location_id)

    count = CorpAsset.objects.filter(
        location_id=location_id,
        location_name__isnull=True
    ).update(location_name_id=location_id)

    count += Clone.objects.filter(
        location_id=location_id,
    ).update(location_name_id=location_id)

    count += JumpClone.objects.filter(
        location_id=location_id,
        location_name__isnull=True
    ).update(location_name_id=location_id)

    count += CharacterMarketOrder.objects.filter(
        location_id=location_id,
        location_name__isnull=True
    ).update(location_name_id=location_id)

    count += Contract.objects.filter(
        start_location_id=location_id,
        start_location_name__isnull=True
    ).update(start_location_name_id=location_id)

    count += CorporateContract.objects.filter(
        start_location_id=location_id,
        start_location_name__isnull=True
    ).update(start_location_name_id=location_id)

    count += Contract.objects.filter(
        end_location_id=location_id,
        end_location_name__isnull=True
    ).update(end_location_name_id=location_id)

    count += CorporateContract.objects.filter(
        end_location_id=location_id,
        end_location_name__isnull=True
    ).update(end_location_name_id=location_id)

    logger.info(f"CT LOCATIONS: Updated {count} models!")
    return count


@shared_task(
    bind=True,
    base=QueueOnce,
    max_retries=None,
    name="corptools.tasks.update_location"
)
@esi_error_retry
def update_location(self, location_id, character_ids=None, force_citadel=False):

    if get_location_cooloff(location_id):
        if force_citadel and location_id > 64000000:
            pass
        else:
            return f"CT LOCATIONS: Cooloff on ID: {location_id}"

    if get_error_count_flag():
        logger.warning(
            f"CT_LOC: Unable to check {location_id} ESI in cooloff...")
        return "ESI Error Limit reached... We will need to try this again later... Somehow..."
        # self.retry(countdown=300)

    if location_id < 64000000:
        location = fetch_location_name(location_id, None, 0, update=True)
        if location is not None:
            location.save()
            return update_missing_locations(location_id)
        else:
            if get_error_count_flag():
                self.retry(countdown=300)

    char_ids = get_character_lists(location_id)

    if len(char_ids) == 0:
        set_location_cooloff(location_id)
        return f"CT LOCATIONS: No more Characters for Location_id: {location_id} cooling off for a while"

    for c_id in char_ids:
        location = fetch_location_name(location_id, None, c_id)
        if location is not None:
            location.save()
            return update_missing_locations(location_id)
        else:
            location_set(location_id, c_id)
            if get_error_count_flag():
                self.retry(countdown=300)

    set_location_cooloff(location_id)
    return f"CT LOCATIONS: No more Characters for Location_id: {location_id} cooling off for a while"


@shared_task(
    bind=True,
    base=QueueOnce,
    name="corptools.tasks.update_all_locations"
)
@no_fail_chain
def update_all_locations(self, character_filter=None, force_citadels=False, update_all=False, chain=[]):
    location_flags = [
        'Deliveries',
        'Hangar',
        'HangarAll'
    ]

    expire = timezone.now() - datetime.timedelta(days=30)  # 1 week refresh

    asset_tops = CharacterAsset.objects.all().values_list("item_id", flat=True)
    char_filter = CharacterAudit.objects.all()

    if character_filter:
        char_filter = char_filter.filter(
            character__character_id__in=character_filter)

    queryset1 = list(
        CharacterAsset.objects.filter(
            location_flag__in=location_flags,
            location_name=None,
            character__in=char_filter
        ).exclude(
            location_id__in=asset_tops
        ).values_list('location_id', flat=True)
    )

    queryset5 = list(
        CharacterAsset.objects.filter(
            location_flag='AssetSafety',
            location_name=None,
            character__in=char_filter
        ).values_list('location_id', flat=True)
    )
    logger.debug(
        f"CT LOCATIONS: {character_filter} Assets {queryset1 + queryset5}")
    queryset3 = list(
        Clone.objects.filter(
            location_id__isnull=False,
            location_name_id__isnull=True,
            character__in=char_filter
        ).values_list('location_id', flat=True)
    )
    logger.debug(f"CT LOCATIONS: {character_filter} Clone {queryset3}")

    queryset4 = list(
        JumpClone.objects.filter(
            location_id__isnull=False,
            location_name_id__isnull=True,
            character__in=char_filter
        ).values_list('location_id', flat=True)
    )
    logger.debug(f"CT LOCATIONS: {character_filter} JumpClone {queryset4}")

    queryset6 = list(
        CharacterMarketOrder.objects.filter(
            location_name_id__isnull=True,
            character__in=char_filter
        ).values_list(
            'location_id',
            flat=True
        )
    )
    logger.debug(
        f"CT LOCATIONS: {character_filter} CharacterMarketOrder {queryset6}")

    queryset7 = list(
        Contract.objects.filter(
            start_location_name=None,
            start_location_id__isnull=False,
            character__in=char_filter
        ).values_list('start_location_id', flat=True)
    )
    logger.debug(f"CT LOCATIONS: {character_filter} Contract {queryset7}")

    queryset8 = list(
        Contract.objects.filter(
            end_location_name=None,
            end_location_id__isnull=False,
            character__in=char_filter
        ).values_list('end_location_id', flat=True)
    )
    logger.debug(f"CT LOCATIONS: {character_filter} Contract {queryset8}")

    all_locations = set(
        queryset1 + queryset3 + queryset8 +
        queryset4 + queryset5 + queryset6 +
        queryset7
    )
    logger.debug(
        f"CT LOCATIONS: {character_filter} all_locations {all_locations}")

    if update_all:
        clones_all = list(
            Clone.objects.all().values_list(
                'location_id',
                flat=True
            )
        )

        jump_clones_all = list(
            Clone.objects.all().values_list(
                'location_id',
                flat=True
            )
        )

        asset_all = list(
            CharacterAsset.objects.all().values_list(
                'location_id',
                flat=True
            )
        )

        queryset2 = list(
            EveLocation.objects.filter(
                last_update__lte=expire,
                location_id__in=set(
                    clones_all + jump_clones_all + asset_all
                )
            ).values_list('location_id', flat=True)
        )

        all_locations += set(queryset2)

    logger.info(f"CT LOCATIONS: {len(all_locations)} Locations to find")

    count = 0
    for location in all_locations:
        if not get_location_cooloff(location):
            update_location.apply_async(
                args=[location],
                kwargs={
                    "force_citadel": force_citadels
                },
                priority=8
            )
            count += 1

    return f"CT LOCATIONS: {character_filter} Sent {count} location_update tasks"



# Generated by Django 4.2.5 on 2023-10-02 11:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('corptools', '0094_lastloginfilter_characteraudit_last_known_login_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='lastloginfilter',
            name='main_corp_only',
            field=models.BooleanField(
                blank=True, default=False, help_text="Only check characters in main's corporation."),
        ),
        migrations.AlterField(
            model_name='lastloginfilter',
            name='days_since_login',
            field=models.IntegerField(
                default=30, help_text='Days since last login of any Character in the accounts before failing the filter.'),
        ),
    ]
